import 'package:basf_logic_components/basf_logic_components.dart';
import 'package:basf_logic_components/domain/models/barcode_id.dart';
import 'package:basf_logic_components/scanner/data/interpreter_data.dart';
import 'package:basf_logic_components/scanner/validators/barcode_utils.dart';
import 'package:basf_logic_components/scanner/validators/regex_helper.dart'
    as regex;
import 'package:gs1_barcode_parser/gs1_barcode_parser.dart';

class ScanInterpreter {
  ScanInterpreter()
      : _barcodeUtils = BarcodeUtils(),
        _data = InterpreterData();

  final InterpreterData _data;
  final BarcodeUtils _barcodeUtils;

  BarcodeData? parseGS1(String barcode) {
    try {
      GS1Barcode gs1barcode = parseBarcode(barcode);

      if (gs1barcode.code.type == CodeType.GS1_128) {
        final String? ssccCode = _barcodeUtils.getSSCCCode(gs1barcode);
        final String? itemCode = _barcodeUtils.getUnitItemCode(gs1barcode);

        if (ssccCode != null || itemCode != null) {
          final BarcodeData barcodeData = BarcodeData(
            barcodeId: ssccCode != null ? BarcodeId.huSSCC : BarcodeId.huItem,
            code: ssccCode ?? itemCode!,
            gs1barcode: gs1barcode,
          );

          return barcodeData;
        }
      }
      return null;
    } catch (_) {
      return null;
    }
  }

  GS1Barcode parseBarcode(String barcode) {
    return _barcodeUtils.hasFNC1symbols(barcode)
        ? _barcodeUtils.parseGS1Barcode(barcode)
        : GS1BarcodeParser.defaultParser().parse(barcode);
  }

  /// multipurpose scanner processing. Can throw
  ScanData getScanData(String barcode) {
    ScanData? data;
    String? errorMessage;

    try {
      // Handle GS1 case
      data = _getDataFromGs1Code(barcode);
    } on Exception catch (_) {
      // Handle non-GS1 QR/2D code
      // 1. TM
      data = _getMaterialScanDataFromQRCode(barcode);
      // 2. Order numbers
      data ??= _getOrderNumbersFromQr(barcode);
      // Handle 1D order number cases:
      data ??= _getOrderNumbersFromBarcode(barcode);
      // Handle plain int numbers.
      data ??= _getIntFromCode(barcode);
    }

    data ??= ScanData(uncategorizedString: barcode);
    return data.copyWith(errorMessage: errorMessage);
  }

  ScanData? _getOrderNumbersFromBarcode(String text) {
    // Barcode:
    // delivery number | delivery position number (ignored?)
    // 3258933524|1
    //
    // Barcode:
    // delivery number  X  order position number
    // 3194084172000010
    final String pattern = '${regex.alternative(
      regex.namedGroup(
        pattern: _data.deliveryNumberPattern,
        name: _data.deliveryKey,
      ),
      regex.namedGroup(
        pattern: _data.stockTransferPattern,
        name: _data.stockTransferKey,
      ),
    )}(?:.${regex.namedGroup(
      pattern: _data.orderItemPattern,
      name: _data.itemKey,
    )})?';

    final RegExpMatch? match = RegExp(pattern).firstMatch(text);
    if (match != null) {
      return ScanData(
        inboundDeliveryNumber: match.namedGroup(_data.deliveryKey),
        outboundDeliveryNumber: match.namedGroup(_data.stockTransferKey),
        orderItemNumber: match.namedGroup(_data.itemKey),
      );
    } else {
      return null;
    }
  }

  ScanData? _getDataFromGs1Code(String? text) {
    final GS1BarcodeParser parser =
        GS1BarcodeParser.configurableParser(GS1BarcodeParserConfig());
    final GS1Barcode result = parser.parse(text ?? '');

    final bool isRecognizedAsGs1 = result.code.type != CodeType.UNDEFINED;
    final bool hasNonGs1FNC1Prefix =
        RegExp(_data.unknownFNC1PrefixPattern).hasMatch(text ?? '');

    String? errorMessage;
    // The GS1 package false-positive case where 10123456789 is interpreted
    // as (10) 123456789
    if (!isRecognizedAsGs1 && !hasNonGs1FNC1Prefix) {
      return null;
    }
    if (!isRecognizedAsGs1) {
      return null;
    }

    final List<GS1ParsedElement> values = result.elements.values.toList();

    return ScanData(
      errorMessage: errorMessage,
      isGS1: true,
      additionalProductId:
          values.valueForExactKey(GS1AICodes.additionalProductId) as String?,
      batch: values.valueForExactKey(GS1AICodes.batchLot) as String?,
      areaSqMeters:
          values.valueForKeyStartingWith(GS1AICodes.areaSqMeters) as double?,
      netWeightKg:
          values.valueForKeyStartingWith(GS1AICodes.netWeightKg) as double?,
      netWeightPounds:
          values.valueForKeyStartingWith(GS1AICodes.netWeightPounds) as double?,
      quantity: values.valueForKeyStartingWith(GS1AICodes.quantity) as double?,
      bestBeforeDate:
          values.valueForExactKey(GS1AICodes.bestBeforeDate) as DateTime?,
      expirationDate:
          values.valueForExactKey(GS1AICodes.expirationDate) as DateTime?,
      dueDate: values.valueForExactKey(GS1AICodes.dueDate) as DateTime?,
      packagingDate:
          values.valueForExactKey(GS1AICodes.packagingDate) as DateTime?,
      productionDate:
          values.valueForExactKey(GS1AICodes.productionDate) as DateTime?,
      depthMeters:
          values.valueForKeyStartingWith(GS1AICodes.depthMeters) as double?,
      lengthMeters:
          values.valueForKeyStartingWith(GS1AICodes.lengthMeters) as double?,
      widthMeters:
          values.valueForKeyStartingWith(GS1AICodes.widthMeters) as double?,
      volumeLitres:
          values.valueForKeyStartingWith(GS1AICodes.volumeLitres) as double?,
      volumeCubicMeters: values
          .valueForKeyStartingWith(GS1AICodes.volumeCubicMeters) as double?,
      variantNumber:
          values.valueForExactKey(GS1AICodes.variantNumber) as String?,
      serialNumber: values.valueForExactKey(GS1AICodes.serialNumber) as String?,
    );
  }

  ScanData? _getOrderNumbersFromQr(String text) {
    // QR-Code / 2D-Code:
    // ignore | ignore | ignore | ignore | ignore | ignore | ignore | order number | order position number | delivery number | might be a delivery position number
    // ScanSol1.0|1|##3|ZHL|Z1E1|ZEL|4948690020|00020|3253753379|2020.07.10
    // Sc01|01|01|Z2L|Z1E1|ZEL|4962477703|00020|3259516252|000001
    final String pattern = regex.namedGroup(
            pattern: _data.orderNumberPattern, name: _data.orderKey) +
        _data.pipePattern +
        regex.namedGroup(
          pattern: _data.orderItemPattern,
          name: _data.itemKey,
        ) +
        _data.pipePattern +
        regex.alternative(
          regex.namedGroup(
            pattern: _data.deliveryNumberPattern,
            name: _data.deliveryKey,
          ),
          regex.namedGroup(
            pattern: _data.stockTransferPattern,
            name: _data.stockTransferKey,
          ),
        ) +
        _data.pipePattern;

    final RegExpMatch? match = RegExp(pattern).firstMatch(text);

    if (match != null) {
      return ScanData(
        orderNumber: match.namedGroup(_data.orderKey),
        inboundDeliveryNumber: match.namedGroup(_data.deliveryKey),
        outboundDeliveryNumber: match.namedGroup(_data.stockTransferKey),
        orderItemNumber: match.namedGroup(_data.itemKey),
      );
    } else {
      return null;
    }
  }

  ScanData? _getMaterialScanDataFromQRCode(String text) {
    final String? plant = _backslashWrappedValueForKey(_data.plantKey, text);
    final String? storageLocation =
        _backslashWrappedValueForKey(_data.storageLocationKey, text);
    final String? equipmentNumber =
        _backslashWrappedValueForKey(_data.equipmentNumberKey, text);
    final String? technicalMaterial =
        _backslashWrappedValueForKey(_data.technicalMaterialKey, text);

    if ((plant?.isNotEmpty ?? false) ||
        (storageLocation?.isNotEmpty ?? false) ||
        (equipmentNumber?.isNotEmpty ?? false) ||
        (technicalMaterial?.isNotEmpty ?? false)) {
      return ScanData(
        plant: plant,
        storageLocation: storageLocation,
        equipmentNumber: equipmentNumber,
        materialNumber: technicalMaterial,
      );
    } else {
      return null;
    }
  }

  String? _backslashWrappedValueForKey(String key, String text) =>
      RegExp(r'\\~' + key + r'(.*?)~\\').firstMatch(text)?.group(1);

  ScanData? _getIntFromCode(String text) {
    final hasDigitsOnly = RegExp(_data.digitsOnlyPattern).hasMatch(text);
    // The condition prevents strings like 'ASK123KDSA' to be treated as '123'
    if (!hasDigitsOnly) {
      return null;
    }
    final int? number = int.tryParse(text);
    return number != null ? ScanData(uncategorizedInt: number) : null;
  }
}
